package bitRanks

import bitRanks.util.Debug

import ap.proof.{ConstraintSimplifier, ExhaustiveProver}
import ap.terfor.{TermOrder, ConstantTerm, OneTerm}
import ap.terfor.conjunctions.Conjunction
import ap.parser.{Parser2InputAbsy, InputAbsy2Internal, Environment,
                  FunctionEncoder, EquivExpander}
import ap.parameters.{GoalSettings, PreprocessingSettings, Param}
import ap.util.FilterIt
import ap.util.{Debug => ApDebug, CmdlParser}

object InteractiveMain {

  private val AC = Debug.AC_MAIN

  //////////////////////////////////////////////////////////////////////////////
  
  Console.withOut(Console.err) {
    printLogo
    println
  }
  
  //////////////////////////////////////////////////////////////////////////////

  private var mainAssertions : Boolean = true    
    
  // We disable the proof tree assertions, because they can take an awful lot of
  // time ...
  ApDebug.enabledAssertions = {
    case (_, ApDebug.AC_PROOF_TREE) => false
    case _ => mainAssertions
  }

  //////////////////////////////////////////////////////////////////////////////

  private val preludeEnv = new Environment
  private val functionEncoder = new FunctionEncoder
  
  private val (backgroundPred, preludeOrder) = Console.withOut(Console.err) {
    print("Reading prelude ... ")
    val reader = ResourceFiles.preludeReader
    val (iBackgroundPred, uniConstants, order) = Parser2InputAbsy(reader, preludeEnv)
    reader.close

    val iBackgroundPred2 = EquivExpander(iBackgroundPred)
    val (iBackgroundPred3, order3) = functionEncoder(iBackgroundPred2, order)
    
    val res = InputAbsy2Internal(functionEncoder.axioms --> iBackgroundPred3, order3)
    functionEncoder.clearAxioms
    
    // we put the (possibly extended) order back into the environment, so that
    // we can continue parsing the transition relations with it
    preludeEnv.order = order3

    println("done")
    (res, order3)
  }
  
  private val operandSorter = new OperandSorter(preludeEnv)
  
  //////////////////////////////////////////////////////////////////////////////
  
  private val ExpansionSettings =
    Param.CONSTRAINT_SIMPLIFIER.set(GoalSettings.DEFAULT,
                                    ConstraintSimplifier.LEMMA_SIMPLIFIER_NON_DNF)
  private val ExpansionProver =
    new ExhaustiveProver(false, ExpansionSettings)
  
  //////////////////////////////////////////////////////////////////////////////

  private def printLogo = {
    println("""    _                                 _           _""") 
    println("""   | |                               | |         | |""")
    println("""    \ \   ____ ____   ____  ___  ____| | _   ____| |""")
    println("""     \ \ / _  )  _ \ / _  )/___)/ ___) || \ / _  | |""")
    println(""" _____) | (/ /| | | ( (/ /|___ ( (___| | | ( ( | | |""")
    println("""(______/ \____)_| |_|\____|___/ \____)_| |_|\_||_|_|""")
  }
  
  def printUsage = {
    println("Usage: seneschal <option>* <inputfile>*")
    println
    printOptions
  }
  
  def printOptions = {
    println("Options:")
    println("  [+-]assert        Enable runtime assertions     (default: +)")
  }

  //////////////////////////////////////////////////////////////////////////////

  def main(args: Array[String]) : Unit = Console.withOut(Console.err) {
    val inputs = try { parseArguments(args) } catch {
      case e : Throwable => {
        println
        println(e.getMessage)
        println
        printUsage
        return
      }
    }

    if (inputs.isEmpty)
      println("No inputs given, exiting")

    for (filename <- inputs) {
      println
      println("Loading file " + filename)
      
      val timeBefore = System.currentTimeMillis
      
      val reader = new java.io.BufferedReader (
                   new java.io.FileReader(new java.io.File (filename)))
      
      computeRanking(reader)
      reader.close
      
      val timeAfter = System.currentTimeMillis
            
      println
      println("" + (timeAfter - timeBefore) + "ms")
      
      /* println
         println(ap.util.Timer)
         ap.util.Timer.reset */
    }
  }

  private def parseArguments(args: Array[String]) : Seq[String] = {
    import CmdlParser._
    
    val inputs = new scala.collection.mutable.ArrayBuffer[String]
    
    for (arg <- args)
      arg match {
        case Opt("assert", value) =>
          mainAssertions = value
        case Opt(_, _) =>
          throw new UnknownArgumentException(arg)
        case _ =>
          inputs += arg
      }
    
    inputs
  }
  
  //////////////////////////////////////////////////////////////////////////////
  
  private def computeRanking(reader : java.io.Reader) : Unit = {
     val (transition, prePostConsts) = parseTransitionRelation(reader)     
     
     println("Expanded transition relation:")
     println(transition)
     
     val finder = new RankingFinder(transition, prePostConsts)
     if (finder.rankingFunction != None) Console.withOut(java.lang.System.out) {
       val funString =
         (for ((c, t) <- finder.rankingFunction.get.elements) yield
           t match {
             case OneTerm => "" + c
             case t : ConstantTerm => "" + c + "*" + 
               ((finder.newConstants get t) match {
                  case None =>
                    t
                  case Some(RankingFinder.NewDivConstant(expr, denom)) =>
                    "(" + expr + " / " + denom + ")"
                })
           }) mkString " + "
       
       println("Ranking function: " + funString)
       println("Lower bound (pre-state): " + finder.preStateLowerBound.get)
       println("Lower bound (post-state): " + finder.postStateLowerBound.get)
     }
  }
  
  //////////////////////////////////////////////////////////////////////////////

  private def parseTransitionRelation(reader : java.io.Reader) = {
    print("Parsing transition relation ... ")
    val env = preludeEnv.clone
    val (preConsts, postConsts, iTransition) = TransitionParser(reader, env)
    println("done")
     
    val order = env.order

    print("Expanding to Presburger formula ... ")

    // we negate the transition relation, because this affects the translation
    // of function applications (and has to be done eventually ...)
    val iTransition2 = operandSorter(!iTransition)
    val iTransition3 = EquivExpander(iTransition2)
    val (iTransition4, order4) = functionEncoder(iTransition3, order)
     
    val transition =
      Conjunction.conj(InputAbsy2Internal(functionEncoder.axioms --> iTransition4,
                                          order4),
                       order4)

    val expansionInput =
      Conjunction.disj(Array(backgroundPred, transition), order4)
    val expandedTransition =
      ExpansionProver(expansionInput, order4).closingConstraint.negate

    println("done")

    (expandedTransition, preConsts zip postConsts)
  }
}
